/* 
 * kniffel.c
 * 
 * Copyright (C) 2006 Felix Bruns (fxb) <felixbruns@web.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 */

#include <stdio.h>
#include <string.h>
#include "pz.h"

/* Kniffel modes */
#define MODE_DICE     (0) /* Roll and select dice */
#define MODE_ENTER    (1) /* Enter score at a position */
#define MODE_FINISHED (2) /* Game is finished */

/* Finished status */
#define FINISHED_ALL   (0) /* The game is finished */
#define FINISHED_UPPER (1) /* Upper fields are finished */
#define FINISHED_LOWER (2) /* Lower fields are finished */

/* Module, window and widget */
static PzModule *module;
static PzWindow *window;
static PzWidget *widget;

/* Global variables */
static int mode, position, dice, tookc, thrown;
static int roll[5], took[5], stat[16], num[6], pasch[5];

/* Text widths & height */
static int widtha, widthb, height;

/* Kniffel field labels */
static const char *labels[] = {
    "Ones", "Twos", "Threes", "Fours",
    "Fives", "Sixes", "Total", "Bonus",
    NULL,
    "3 of a kind", "4 of a kind", "Full House", "S.Straight",
    "L.Straight", "Kniffel", "Chance", "Total",
    NULL
};

/*
  Funtion:      kniffel_finished
  Description:  Checks the current status of the game.
  Parameters:   int c | Status code to check.
  Return value: 0 if status matches, 1 if not.
*/
static int kniffel_finished(int c){
    int i;
    
    switch(c){
        case FINISHED_ALL:
            return (kniffel_finished(FINISHED_UPPER) && kniffel_finished(FINISHED_LOWER));
            break;
        case FINISHED_UPPER:
            for(i=0; i<6; i++){if(stat[i] == -1) return 0;}
            break;
        case FINISHED_LOWER:
            for(i=8; i<15; i++){if(stat[i] == -1) return 0;}
            break;
    }
    
    return 1;
}

/*
  Funtion:      kniffel_rearrange
  Description:  Rearranges the gobal integer array 'roll' after a dice was taken.
  Parameters:   None.
  Return value: None.
*/
static void kniffel_rearrange(){
    int i, j, _roll[5];
    
    for(i=0, j=0; i<5; i++){
        if(roll[i] != -1){
            _roll[j] = roll[i];
            j++;
        }
    }
    
    memcpy(roll, _roll, 5*sizeof(int));
}

/*
  Funtion:      kniffel_roll
  Description:  Rolls the dice. Fills the gobal integer array 'roll' with random values from 1 to 6.
  Parameters:   None.
  Return value: None.
*/
static void kniffel_roll(){
    int i;
    
    for(i=0; i<5-tookc; i++){
        roll[i] = rand()%6+1;
    }
}

/*
  Funtion:      kniffel_reset_roll
  Description:  Resets global variables before a new roll.
  Parameters:   None.
  Return value: None.
*/
static void kniffel_reset_roll(){
    int i = 0;

    memset(roll, -1, 5*sizeof(int));
    memset(took, -1, 5*sizeof(int));
    memset(num,   0, 6*sizeof(int));
    memset(pasch, 0, 5*sizeof(int));
    position = 0;
    while(stat[position] != -1 ||
          position == 6 ||
          position == 7 ||
          position == 15){
        if(i > 15) break;
        position++;
        if(position > 15) position = 0;
        i++;
    }
    dice     = 0;
    tookc    = 0;
    thrown   = 1;
}

/*
  Funtion:      kniffel_reset
  Description:  Resets all global variables before a new game.
  Parameters:   None.
  Return value: None.
*/
static void kniffel_reset(){
    memset(stat, -1, 16*sizeof(int));
    mode = MODE_DICE;
    kniffel_reset_roll();    
    kniffel_roll();
}

/*
  Funtion:      kniffel_calculate
  Description:  Calculates the score for the current roll and the selected field.
                Also checks for pairs, triplets, quadruplets, kniffels and straights.
  Parameters:   None.
  Return value: Score for the current roll at the current position.
*/
static int kniffel_calculate(){
    int i, j, val = 0, sum = 0;
    
    /* Calculate pips of all dice together */
    for(i=0; i<5; i++){
        sum += took[i];
    }
    
    /* Calculates the number of times a dice (e.g. 4) exist in the current roll.
       Checks also if there are pairs, triplets, quadruplets, kniffels or straights. */
    for(i=0; i<6; i++){
       for(j=0; j<5; j++){
           if(took[j] == i+1){
               num[i]++;
           }
       }
       if(num[i] > 1){
          pasch[num[i]-1] = 1;
          pasch[0]        = 1;
       }
    }
    
    switch(position){
        /* Ones, Twos, Threes, Fours, Fives, Sixes */
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
            for(i=0; i<5; i++){
                val += (took[i] == position+1)?(position+1):0;
            }
            break;
        /* 3 of a kind */
        case 8:
            if(pasch[2] || pasch[3] || pasch[4]){
                val = sum;
            }
            break;
        /* 4 of a kind */
        case 9:
            if(pasch[3] || pasch[4]){
                val = sum;
            }
            break;
        /* Full House */
        case 10:
            if((pasch[1] && pasch[2]) || pasch[4]){
                val = 25;
            }            
            break;
        /* Straights */
        case 11:
        case 12:
                i = 0;
                for(j=1; j<=5; j++){
                        if(num[j] > 0 && num[j-1] > 0){
                            i++;
                        }
                        else if(i > 0 && i < 3){
                            i = 0;
                        }
                }
                if(i == 4 && position == 12){
                    val = 40;
                }
                if(i >= 3 && position == 11){
                    val = 30;
                }
            break;
        /* Kniffel! Yay! */
        case 13:
            if(pasch[4]){
                val = 50;
            }            
            break;
        /* Chance */
        case 14:
            val = sum;
            break;
    }
    
    return val;
}

/*
  Funtion:      kniffel_draw_dice
  Description:  Draws a dice.
  Parameters:   ttk_surface srf | Surface to draw to.
                int x           | X-Location of the dice.
                int y           | Y-Location of the dice.
                int n           | Pips of the dice.
                float scale     | Scale by this value.
                int selected    | Draw a selected dice (other color).
  Return value: None.
*/
static void kniffel_draw_dice(ttk_surface srf, int x, int y, int n, float scale, int selected)
{
    ttk_color col = (selected)?(ttk_ap_getx("box.selected.fg")->color):(ttk_ap_getx("box.default.fg")->color);
    int s2  = 2*scale;
    int s3  = 3*scale;
    int s9  = 9*scale;
    int s16 = 16*scale;

    ttk_ap_fillrect(srf, (selected)?ttk_ap_getx("box.selected.bg"):ttk_ap_getx("box.default.bg"),
                    x, y, x+20*scale, y+20*scale);
    if(n == 1 || n == 3 || n == 5){
        ttk_aafillellipse(srf, x+s9, y+s9, s2, s2, col);
    }
    if(n == 2 || n == 3 || n == 4 || n == 5 || n == 6){
        ttk_aafillellipse(srf, x+s3,  y+ s3, s2, s2, col);
        ttk_aafillellipse(srf, x+s16, y+s16, s2, s2, col);
    }
    if(n == 4 || n == 5 || n == 6){
        ttk_aafillellipse(srf, x+s16, y+ s3, s2, s2, col);
        ttk_aafillellipse(srf, x+s3,  y+s16, s2, s2, col);
    }
    if(n == 6){
        ttk_aafillellipse(srf, x+s16, y+s9, s2, s2, col);
        ttk_aafillellipse(srf, x+s3,  y+s9, s2, s2, col);
    }
}

/*
  Funtion:      kniffel_event_draw
  Description:  Draws lines, selector, field labels, values, dice, taken dice and score to the main surface.
  Parameters:   TWidget *wid    | Pointer to a TWidget structure.
                ttk_surface srf | Surface to draw to.
  Return value: None.
*/
static void kniffel_event_draw(TWidget *wid, ttk_surface srf)
{
    ttk_color col = ttk_ap_getx("window.fg")->color;
    float scale = (float)ttk_screen->h/132;
    char val[3];
    
    /* Clear screen */
    ttk_ap_fillrect(srf, ttk_ap_getx("window.bg"), 0, 0, ttk_screen->w, ttk_screen->h);
    
    /* Draw lines */
    ttk_line(srf, ttk_screen->w/2-15, 0, ttk_screen->w/2-15, 8*height+2, col);
    ttk_line(srf, widtha+5, 0, widtha+5, 8*height+2, col);
    ttk_line(srf, ttk_screen->w/2-5+widthb, 0, ttk_screen->w/2-5+widthb, 8*height+2, col);
    ttk_line(srf, 0, 8*height+2, ttk_screen->w, 8*height+2, col);
    
    /* Draw selector */
    if(position < 8){
        ttk_ap_fillrect(srf, ttk_ap_getx("box.selected.bg"), 0, position*height,
                     5+widtha, (position+1)*height);
    }
    else{
        ttk_ap_fillrect(srf, ttk_ap_getx("box.selected.bg"),ttk_screen->w/2-14, (position-8)*height,
                     ttk_screen->w/2-5+widthb, (position-7)*height);
    }
    
    /* Draw field labels */
    int i = 0;
    while(labels[i] != NULL){
        ttk_text(srf, ttk_textfont, 2, i*height,
                 (position==i)?(ttk_ap_getx("box.selected.fg")->color):col, labels[i]);
        i++;
    }
    i++;
    while(labels[i] != NULL){
        ttk_text(srf, ttk_textfont, ttk_screen->w/2-10, (i-9)*height,
                 (position==i-1)?(ttk_ap_getx("box.selected.fg")->color):col, labels[i]);
        i++;
    }
    
    /* Draw values */
    for(i=0; i<8; i++){
        if(stat[i] != -1){
            sprintf(val, "%d", stat[i]);
            ttk_text(srf, ttk_textfont, widtha+8, i*height, col, val);
        }
    }
    for(i=8; i<16; i++){
        if(stat[i] != -1){
            sprintf(val, "%d", stat[i]);
            ttk_text(srf, ttk_textfont, widthb+ttk_screen->w/2-2, (i-8)*height, col, val);
        }
    }
    
    /* Draw dice */
    if(mode == MODE_DICE){
        for(i=0; i<5-tookc; i++){
            kniffel_draw_dice(srf, (i*21+2)*scale, ttk_screen->h-22*scale+2, roll[i], scale, (dice==i));
        }
    }
    
    /* Draw taken dice */
    for(i=0; i<tookc; i++){
        kniffel_draw_dice(srf, ttk_screen->w-((i+1)*21+2)*scale, ttk_screen->h-22*scale+2, took[i], scale, 0);
    }
    
    /* Draw score */
    if(mode == MODE_FINISHED){
        int width;
        char buffer[40];
        sprintf(buffer, "You reached a total of %d!", stat[6] + stat[7] + stat[15]);
        width = ttk_text_width(ttk_textfont, buffer);
        ttk_text(srf, ttk_textfont, ttk_screen->w/2-width/2, ttk_screen->h-15*scale, col, buffer);
    }
}

/*
  Funtion:      kniffel_event_scroll
  Description:  Handles scroll events.
  Parameters:   TWidget *wid    | Pointer to a TWidget structure.
                int dir         | Distance and direction scrolled.
  Return value: TTK event code.
*/
static int kniffel_event_scroll(TWidget *wid, int dir)
{
    int i;
#ifdef IPOD
    TTK_SCROLLMOD(dir, 7);
#endif
    switch(mode){
        /* Change mode and either change current field or current dice */
        case MODE_ENTER:
            if(dir < 0){
                position--;
                if(position < 0) position = 15;
                i = 0;
                while(stat[position] != -1 ||
                      position == 6 ||
                      position == 7 ||
                      position == 15){
                    if(i > 15) break;
                    position--;
                    if(position < 0) position = 15;
                    i++;
                }
            }
            else{
                position++;
                if(position > 15) position = 0;
                i = 0;
                while(stat[position] != -1 ||
                      position == 6 ||
                      position == 7 ||
                      position == 15){
                    if(i > 15) break;
                    position++;
                    if(position > 15) position = 0;
                    i++;
                }
            }
            break;
        case MODE_DICE:
            if(dir < 0){
                dice--;
                if(dice < 0) dice = 4-tookc;
            }
            else{
                dice++;
                if(dice > 4-tookc) dice = 0;
            }
            break;
    }
    wid->dirty++;
    return TTK_EV_CLICK;
}

/*
  Funtion:      kniffel_event_down
  Description:  Handles button press events.
  Parameters:   TWidget *wid    | Pointer to a TWidget structure.
                int button      | Button pressed.
  Return value: TTK event code.
*/
static int kniffel_event_down(TWidget *wid, int button)
{
    int i;
    
    /* If finished press any key to start a new game */
    if(mode == MODE_FINISHED){
        kniffel_reset();
        wid->dirty++;
        return TTK_EV_UNUSED;
    }
    
    switch(button){
        case TTK_BUTTON_ACTION:
            /* Check mode and either take a dice or select a field and save the score.
               If all dice are gone goto enter mode */
            if(mode == MODE_DICE){
                took[tookc] = roll[dice];
                roll[dice]  = -1;
                dice        = 0;
                kniffel_rearrange();
                tookc++;
                if(tookc == 5){
                    mode = MODE_ENTER;
                }
            }
            else{
                stat[position] = kniffel_calculate();
                if(kniffel_finished(FINISHED_UPPER)){
                    stat[6]  = stat[0] + stat[1] + stat[2];
                    stat[6] += stat[3] + stat[4] + stat[5];
                    stat[7] = (stat[6] >= 63)?35:0;
                }
                if(kniffel_finished(FINISHED_LOWER)){
                    stat[15]  = stat[8] + stat[9] + stat[10];
                    stat[15] += stat[11] + stat[12] + stat[13];
                    stat[15] += stat[14];
                }
                if(!kniffel_finished(FINISHED_ALL)){
                    kniffel_reset_roll();
                    kniffel_roll();
                    mode = MODE_DICE;
                }
                else{
                    tookc = 0;
                    mode = MODE_FINISHED;
                }
            }
            wid->dirty++;
            break;
        case TTK_BUTTON_PLAY:
            /* Roll the dice. If it's the 3rd throw goto enter mode */
            if(mode == MODE_DICE){
                kniffel_roll();
                thrown++;
                if(thrown == 3){
                    for(i=0; tookc<5; i++, tookc++){
                        took[tookc] = roll[i];
                        roll[i]     = -1;
                    }
                    mode = MODE_ENTER;
                }
            }
            wid->dirty++;
            break;
        case TTK_BUTTON_MENU:
            /* Bye bye Kniffel player! */
            pz_close_window(wid->win);
            break;
        default:
            return TTK_EV_UNUSED;
            break;
    }
    
    return TTK_EV_UNUSED;
}

/*
  Funtion:      new_kniffel_window
  Description:  Opens a new Kniffel window and starts a game.
  Parameters:   None.
  Return value: A pointer to a PzWindow structure.
*/
static PzWindow *new_kniffel_window()
{
    /* Create module window and set callback function */
    window            = pz_new_window(_("Kniffel"), PZ_WINDOW_NORMAL);
    widget            = ttk_new_widget(0, 0);
    widget->w         = ttk_screen->w;
    widget->h         = ttk_screen->h;
    widget->draw      = kniffel_event_draw;
    widget->down      = kniffel_event_down;
    widget->scroll    = kniffel_event_scroll;
    widget->focusable = 1;
    
    /* Get text widths & height to calculate drawing positions later */
    widtha = ttk_text_width(ttk_textfont, labels[2]);
    widthb = ttk_text_width(ttk_textfont, labels[9]);
    height = ttk_text_height(ttk_textfont);
    
    /* Initialize global variables */
    kniffel_reset();
    
    /* Hide the header and add main widget to module window */
    ttk_window_hide_header(window);
    ttk_add_widget(window, widget);
    
    return pz_finish_window(window);
}

/*
  Funtion:      kniffel_init
  Description:  Registers the module and adds its menu item.
  Parameters:   None.
  Return value: None.
*/
static void kniffel_init()
{
    module = pz_register_module("kniffel", NULL);
    pz_menu_add_action("/Extras/Games/Kniffel", new_kniffel_window);
}

/* Init */
PZ_MOD_INIT(kniffel_init)
